@@ -9,18 +9,25 @@
 
 {-# Language CPP #-}
 {-# Language TemplateHaskell #-}
+{-# Language TypeOperators #-}
 -- Adapted from https://wiki.haskell.org/A_practical_Template_Haskell_Tutorial
 
 module Rank2.TH (deriveAll, deriveFunctor, deriveApply, unsafeDeriveApply, deriveApplicative,
-                 deriveFoldable, deriveTraversable, deriveDistributive, deriveDistributiveTraversable)
+                 deriveFoldable, deriveTraversable,
+                 deriveDistributive, deriveDistributiveTraversable, deriveLogistic)
 where
 
 import Control.Applicative (liftA2, liftA3)
 import Control.Monad (replicateM)
+import Data.Bifunctor (first)
 import Data.Distributive (cotraverse)
-import Data.Monoid ((<>))
-import Language.Haskell.TH
-import Language.Haskell.TH.Syntax (BangType, VarBangType, getQ, putQ)
+import Data.Functor.Compose (Compose (Compose))
+import Data.Functor.Contravariant (Contravariant, contramap)
+import qualified Language.Haskell.TH as TH
+import Language.Haskell.TH (Q, TypeQ, Name, TyVarBndr(KindedTV, PlainTV), Clause, Dec(..), Con(..), Type(..), Exp(..),
+                            Inline(Inlinable, Inline), RuleMatch(FunLike), Phases(AllPhases),
+                            appE, conE, conP, conT, instanceD, varE, varP, varT, normalB, pragInlD, recConE, wildP)
+import Language.Haskell.TH.Syntax (BangType, VarBangType, Info(TyConI), getQ, putQ, newName)
 
 import qualified Rank2
 
@@ -28,21 +35,22 @@ data Deriving = Deriving { _derivingConstructor :: Name, _derivingVariable :: Na
 
 deriveAll :: Name -> Q [Dec]
 deriveAll ty = foldr f (pure []) [deriveFunctor, deriveApply, deriveApplicative,
-                                  deriveFoldable, deriveTraversable, deriveDistributive, deriveDistributiveTraversable]
+                                  deriveFoldable, deriveTraversable,
+                                  deriveDistributive, deriveDistributiveTraversable, deriveLogistic]
    where f derive rest = (<>) <$> derive ty <*> rest
 
 deriveFunctor :: Name -> Q [Dec]
 deriveFunctor ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Functor ty
-   (constraints, dec) <- genFmap cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   (constraints, dec) <- genFmap instanceType cs
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, pragInlD '(Rank2.<$>) Inline FunLike AllPhases]]
 
 deriveApply :: Name -> Q [Dec]
 deriveApply ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Apply ty
-   (constraints, dec) <- genAp cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   (constraints, dec) <- genAp instanceType cs
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, genLiftA2 cs, genLiftA3 cs,
               pragInlD '(Rank2.<*>) Inlinable FunLike AllPhases,
               pragInlD 'Rank2.liftA2 Inlinable FunLike AllPhases]]
@@ -51,8 +59,8 @@ deriveApply ty = do
 unsafeDeriveApply :: Name -> Q [Dec]
 unsafeDeriveApply ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Apply ty
-   (constraints, dec) <- genApUnsafely cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   (constraints, dec) <- genApUnsafely instanceType cs
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, genLiftA2Unsafely cs, genLiftA3Unsafely cs,
               pragInlD '(Rank2.<*>) Inlinable FunLike AllPhases,
               pragInlD 'Rank2.liftA2 Inlinable FunLike AllPhases]]
@@ -61,39 +69,46 @@ deriveApplicative :: Name -> Q [Dec]
 deriveApplicative ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Applicative ty
    (constraints, dec) <- genPure cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, pragInlD 'Rank2.pure Inline FunLike AllPhases]]
 
 deriveFoldable :: Name -> Q [Dec]
 deriveFoldable ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Foldable ty
-   (constraints, dec) <- genFoldMap cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   (constraints, dec) <- genFoldMap instanceType cs
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, pragInlD 'Rank2.foldMap Inlinable FunLike AllPhases]]
 
 deriveTraversable :: Name -> Q [Dec]
 deriveTraversable ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Traversable ty
-   (constraints, dec) <- genTraverse cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   (constraints, dec) <- genTraverse instanceType cs
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, pragInlD 'Rank2.traverse Inlinable FunLike AllPhases]]
 
 deriveDistributive :: Name -> Q [Dec]
 deriveDistributive ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Distributive ty
    (constraints, dec) <- genCotraverse cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
              [pure dec, pragInlD 'Rank2.cotraverse Inline FunLike AllPhases]]
 
 deriveDistributiveTraversable :: Name -> Q [Dec]
 deriveDistributiveTraversable ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.DistributiveTraversable ty
    (constraints, dec) <- genCotraverseTraversable cs
-   sequence [instanceD (cxt $ map pure constraints) instanceType [pure dec]]
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType [pure dec]]
+
+deriveLogistic :: Name -> Q [Dec]
+deriveLogistic ty = do
+   (instanceType, cs) <- reifyConstructors ''Rank2.Logistic ty
+   (constraints, decs) <- genDeliver instanceType cs
+   sequence [instanceD (TH.cxt $ map pure constraints) instanceType
+              (map pure decs <> [pragInlD 'Rank2.deliver Inline FunLike AllPhases])]
 
 reifyConstructors :: Name -> Name -> Q (TypeQ, [Con])
 reifyConstructors cls ty = do
-   (TyConI tyCon) <- reify ty
+   (TyConI tyCon) <- TH.reify ty
    (tyConName, tyVars, _kind, cs) <- case tyCon of
       DataD _ nm tyVars kind cs _   -> return (nm, tyVars, kind, cs)
       NewtypeD _ nm tyVars kind c _ -> return (nm, tyVars, kind, [c])
@@ -101,54 +116,64 @@ reifyConstructors cls ty = do
  
 #if MIN_VERSION_template_haskell(2,17,0)
    let (KindedTV tyVar () (AppT (AppT ArrowT _) StarT)) = last tyVars
-       instanceType           = conT cls `appT` foldl apply (conT tyConName) (init tyVars)
-       apply t (PlainTV name _)    = appT t (varT name)
-       apply t (KindedTV name _ _) = appT t (varT name)
+       instanceType           = conT cls `TH.appT` foldl apply (conT tyConName) (init tyVars)
+       apply t (PlainTV name _)    = TH.appT t (varT name)
+       apply t (KindedTV name _ _) = TH.appT t (varT name)
 #else
    let (KindedTV tyVar (AppT (AppT ArrowT _) StarT)) = last tyVars
-       instanceType           = conT cls `appT` foldl apply (conT tyConName) (init tyVars)
-       apply t (PlainTV name)    = appT t (varT name)
-       apply t (KindedTV name _) = appT t (varT name)
+       instanceType           = conT cls `TH.appT` foldl apply (conT tyConName) (init tyVars)
+       apply t (PlainTV name)    = TH.appT t (varT name)
+       apply t (KindedTV name _) = TH.appT t (varT name)
 #endif
  
    putQ (Deriving tyConName tyVar)
    return (instanceType, cs)
 
-genFmap :: [Con] -> Q ([Type], Dec)
-genFmap cs = do (constraints, clauses) <- unzip <$> mapM genFmapClause cs
-                return (concat constraints, FunD '(Rank2.<$>) clauses)
+genFmap :: TypeQ -> [Con] -> Q ([Type], Dec)
+genFmap instanceType cs = do
+   it <- instanceType
+   (constraints, clauses) <- unzip <$> mapM (genFmapClause it) cs
+   return (concat constraints, FunD '(Rank2.<$>) clauses)
 
-genAp :: [Con] -> Q ([Type], Dec)
-genAp [con] = do (constraints, clause) <- genApClause False con
-                 return (constraints, FunD '(Rank2.<*>) [clause])
+genAp :: TypeQ -> [Con] -> Q ([Type], Dec)
+genAp instanceType [con] = do
+   it <- instanceType
+   (constraints, clause) <- genApClause False it con
+   return (constraints, FunD '(Rank2.<*>) [clause])
 
 genLiftA2 :: [Con] -> Q Dec
-genLiftA2 [con] = funD 'Rank2.liftA2 [genLiftA2Clause False con]
+genLiftA2 [con] = TH.funD 'Rank2.liftA2 [genLiftA2Clause False con]
 
 genLiftA3 :: [Con] -> Q Dec
-genLiftA3 [con] = funD 'Rank2.liftA3 [genLiftA3Clause False con]
+genLiftA3 [con] = TH.funD 'Rank2.liftA3 [genLiftA3Clause False con]
 
-genApUnsafely :: [Con] -> Q ([Type], Dec)
-genApUnsafely cons = do (constraints, clauses) <- unzip <$> mapM (genApClause True) cons
-                        return (concat constraints, FunD '(Rank2.<*>) clauses)
+genApUnsafely :: TypeQ -> [Con] -> Q ([Type], Dec)
+genApUnsafely instanceType cons = do
+   it <- instanceType
+   (constraints, clauses) <- unzip <$> mapM (genApClause True it) cons
+   return (concat constraints, FunD '(Rank2.<*>) clauses)
 
 genLiftA2Unsafely :: [Con] -> Q Dec
-genLiftA2Unsafely cons = funD 'Rank2.liftA2 (genLiftA2Clause True <$> cons)
+genLiftA2Unsafely cons = TH.funD 'Rank2.liftA2 (genLiftA2Clause True <$> cons)
 
 genLiftA3Unsafely :: [Con] -> Q Dec
-genLiftA3Unsafely cons = funD 'Rank2.liftA3 (genLiftA3Clause True <$> cons)
+genLiftA3Unsafely cons = TH.funD 'Rank2.liftA3 (genLiftA3Clause True <$> cons)
 
 genPure :: [Con] -> Q ([Type], Dec)
 genPure cs = do (constraints, clauses) <- unzip <$> mapM genPureClause cs
                 return (concat constraints, FunD 'Rank2.pure clauses)
 
-genFoldMap :: [Con] -> Q ([Type], Dec)
-genFoldMap cs = do (constraints, clauses) <- unzip <$> mapM genFoldMapClause cs
-                   return (concat constraints, FunD 'Rank2.foldMap clauses)
+genFoldMap :: TypeQ -> [Con] -> Q ([Type], Dec)
+genFoldMap instanceType cs = do
+   it <- instanceType
+   (constraints, clauses) <- unzip <$> mapM (genFoldMapClause it) cs
+   return (concat constraints, FunD 'Rank2.foldMap clauses)
 
-genTraverse :: [Con] -> Q ([Type], Dec)
-genTraverse cs = do (constraints, clauses) <- unzip <$> mapM genTraverseClause cs
-                    return (concat constraints, FunD 'Rank2.traverse clauses)
+genTraverse :: TypeQ -> [Con] -> Q ([Type], Dec)
+genTraverse instanceType cs = do
+   it <- instanceType
+   (constraints, clauses) <- unzip <$> mapM (genTraverseClause it) cs
+   return (concat constraints, FunD 'Rank2.traverse clauses)
 
 genCotraverse :: [Con] -> Q ([Type], Dec)
 genCotraverse [con] = do (constraints, clause) <- genCotraverseClause con
@@ -158,19 +183,40 @@ genCotraverseTraversable :: [Con] -> Q ([Type], Dec)
 genCotraverseTraversable [con] = do (constraints, clause) <- genCotraverseTraversableClause con
                                     return (constraints, FunD 'Rank2.cotraverseTraversable [clause])
 
-genFmapClause :: Con -> Q ([Type], Clause)
-genFmapClause (NormalC name fieldTypes) = do
+genDeliver :: TypeQ -> [Con] -> Q ([Type], [Dec])
+genDeliver instanceType [con] = do
+   it <- instanceType
+   let AppT _classType rt = it
+       recType = pure rt
+   signable <- TH.isExtEnabled TH.InstanceSigs
+   scopable <- TH.isExtEnabled TH.ScopedTypeVariables
+   if signable && scopable then do
+      p <- newName "p"
+      q <- newName "q"
+      (constraints, clause) <- genDeliverClause recType (Just q) con
+      ctx <- [t| Contravariant $(varT p) |]
+      methodType <- [t| $(varT p) ($(recType) $(varT q) -> $(recType) $(varT q)) -> $(recType) (Compose $(varT p) ($(varT q) Rank2.~> $(varT q))) |]
+      return (constraints,
+              [SigD 'Rank2.deliver (ForallT [binder p, binder q] [ctx] methodType),
+               FunD 'Rank2.deliver [clause]])
+   else do
+      (constraints, clause) <- genDeliverClause recType Nothing con
+      return (constraints, [FunD 'Rank2.deliver [clause]])
+
+
+genFmapClause :: Type -> Con -> Q ([Type], Clause)
+genFmapClause _ (NormalC name fieldTypes) = do
    f          <- newName "f"
    fieldNames <- replicateM (length fieldTypes) (newName "x")
    let pats = [varP f, conP name (map varP fieldNames)]
        constraintsAndFields = zipWith newField fieldNames fieldTypes
        newFields = map (snd <$>) constraintsAndFields
-       body = normalB $ appsE $ conE name : newFields
+       body = normalB $ TH.appsE $ conE name : newFields
        newField :: Name -> BangType -> Q ([Type], Exp)
        newField x (_, fieldType) = genFmapField (varE f) fieldType (varE x) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause pats body []
-genFmapClause (RecC name fields) = do
+   (,) constraints <$> TH.clause pats body []
+genFmapClause _ (RecC name fields) = do
    f <- newName "f"
    x <- newName "x"
    let body = normalB $ recConE name $ (snd <$>) <$> constraintsAndFields
@@ -178,18 +224,20 @@ genFmapClause (RecC name fields) = do
        newNamedField :: VarBangType -> Q ([Type], (Name, Exp))
        newNamedField (fieldName, _, fieldType) =
           ((,) fieldName <$>)
-          <$> genFmapField (varE f) fieldType (appE (varE fieldName) (varE x)) id
+          <$> genFmapField (varE f) fieldType (getFieldOf x fieldName) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP f, x `asP` recP name []] body []
-genFmapClause (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
+   (,) constraints <$> TH.clause [varP f, x `TH.asP` TH.recP name []] body []
+genFmapClause instanceType (GadtC [name] fieldTypes _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genFmapClause (NormalC name fieldTypes)
-genFmapClause (RecGadtC [name] fields _resultType@(AppT _ (VarT tyVar))) =
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genFmapClause instanceType (NormalC name fieldTypes)
+genFmapClause instanceType (RecGadtC [name] fields _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genFmapClause (RecC name fields)
-genFmapClause (ForallC _vars _cxt con) = genFmapClause con
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genFmapClause instanceType (RecC name fields)
+genFmapClause instanceType (ForallC _vars _cxt con) = genFmapClause instanceType con
 
 genFmapField :: Q Exp -> Type -> Q Exp -> (Q Exp -> Q Exp) -> Q ([Type], Exp)
 genFmapField fun fieldType fieldAccess wrap = do
@@ -209,10 +257,10 @@ genLiftA2Clause unsafely (NormalC name fieldTypes) = do
    y <- newName "y"
    fieldNames2 <- replicateM (length fieldTypes) (newName "y")
    let pats = [varP f, conP name (map varP fieldNames1), varP y]
-       body = normalB $ appsE $ conE name : zipWith newField (zip fieldNames1 fieldNames2) fieldTypes
+       body = normalB $ TH.appsE $ conE name : zipWith newField (zip fieldNames1 fieldNames2) fieldTypes
        newField :: (Name, Name) -> BangType -> Q Exp
        newField (x, y) (_, fieldType) = genLiftA2Field unsafely (varE f) fieldType (varE x) (varE y) id
-   clause pats body [valD (conP name $ map varP fieldNames2) (normalB $ varE y) []]
+   TH.clause pats body [TH.valD (conP name $ map varP fieldNames2) (normalB $ varE y) []]
 genLiftA2Clause unsafely (RecC name fields) = do
    f <- newName "f"
    x <- newName "x"
@@ -220,9 +268,9 @@ genLiftA2Clause unsafely (RecC name fields) = do
    let body = normalB $ recConE name $ map newNamedField fields
        newNamedField :: VarBangType -> Q (Name, Exp)
        newNamedField (fieldName, _, fieldType) =
-          fieldExp fieldName (genLiftA2Field unsafely (varE f) fieldType (getFieldOf x) (getFieldOf y) id)
-          where getFieldOf = appE (varE fieldName) . varE
-   clause [varP f, x `asP` recP name [], varP y] body []
+          TH.fieldExp fieldName $
+             genLiftA2Field unsafely (varE f) fieldType (getFieldOf x fieldName) (getFieldOf y fieldName) id
+   TH.clause [varP f, x `TH.asP` TH.recP name [], varP y] body []
 genLiftA2Clause unsafely (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -255,11 +303,11 @@ genLiftA3Clause unsafely (NormalC name fieldTypes) = do
    fieldNames2 <- replicateM (length fieldTypes) (newName "y")
    fieldNames3 <- replicateM (length fieldTypes) (newName "z")
    let pats = [varP f, conP name (map varP fieldNames1), varP y, varP z]
-       body = normalB $ appsE $ conE name : zipWith newField (zip3 fieldNames1 fieldNames2 fieldNames3) fieldTypes
+       body = normalB $ TH.appsE $ conE name : zipWith newField (zip3 fieldNames1 fieldNames2 fieldNames3) fieldTypes
        newField :: (Name, Name, Name) -> BangType -> Q Exp
        newField (x, y, z) (_, fieldType) = genLiftA3Field unsafely (varE f) fieldType (varE x) (varE y) (varE z) id
-   clause pats body [valD (conP name $ map varP fieldNames2) (normalB $ varE y) [],
-                     valD (conP name $ map varP fieldNames3) (normalB $ varE z) []]
+   TH.clause pats body [TH.valD (conP name $ map varP fieldNames2) (normalB $ varE y) [],
+                        TH.valD (conP name $ map varP fieldNames3) (normalB $ varE z) []]
 genLiftA3Clause unsafely (RecC name fields) = do
    f <- newName "f"
    x <- newName "x"
@@ -268,9 +316,9 @@ genLiftA3Clause unsafely (RecC name fields) = do
    let body = normalB $ recConE name $ map newNamedField fields
        newNamedField :: VarBangType -> Q (Name, Exp)
        newNamedField (fieldName, _, fieldType) =
-          fieldExp fieldName (genLiftA3Field unsafely (varE f) fieldType (getFieldOf x) (getFieldOf y) (getFieldOf z) id)
-          where getFieldOf = appE (varE fieldName) . varE
-   clause [varP f, x `asP` recP name [], varP y, varP z] body []
+          TH.fieldExp fieldName
+             (genLiftA3Field unsafely (varE f) fieldType (getFieldOf x fieldName) (getFieldOf y fieldName) (getFieldOf z fieldName) id)
+   TH.clause [varP f, x `TH.asP` TH.recP name [], varP y, varP z] body []
 genLiftA3Clause unsafely (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -297,39 +345,40 @@ genLiftA3Field unsafely fun fieldType field1Access field2Access field3Access wra
      _ | unsafely -> field1Access
        | otherwise -> error ("Cannot apply liftA3 to field of type " <> show fieldType)
 
-genApClause :: Bool -> Con -> Q ([Type], Clause)
-genApClause unsafely (NormalC name fieldTypes) = do
+genApClause :: Bool -> Type -> Con -> Q ([Type], Clause)
+genApClause unsafely _ (NormalC name fieldTypes) = do
    fieldNames1 <- replicateM (length fieldTypes) (newName "x")
    fieldNames2 <- replicateM (length fieldTypes) (newName "y")
    rhsName <- newName "rhs"
    let pats = [conP name (map varP fieldNames1), varP rhsName]
        constraintsAndFields = zipWith newField (zip fieldNames1 fieldNames2) fieldTypes
        newFields = map (snd <$>) constraintsAndFields
-       body = normalB $ appsE $ conE name : newFields
+       body = normalB $ TH.appsE $ conE name : newFields
        newField :: (Name, Name) -> BangType -> Q ([Type], Exp)
        newField (x, y) (_, fieldType) = genApField unsafely fieldType (varE x) (varE y) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause pats body [valD (conP name $ map varP fieldNames2) (normalB $ varE rhsName) []]
-genApClause unsafely (RecC name fields) = do
+   (,) constraints <$> TH.clause pats body [TH.valD (conP name $ map varP fieldNames2) (normalB $ varE rhsName) []]
+genApClause unsafely _ (RecC name fields) = do
    x <- newName "x"
    y <- newName "y"
    let body = normalB $ recConE name $ (snd <$>) <$> constraintsAndFields
        constraintsAndFields = map newNamedField fields
        newNamedField :: VarBangType -> Q ([Type], (Name, Exp))
        newNamedField (fieldName, _, fieldType) =
-          ((,) fieldName <$>) <$> genApField unsafely fieldType (getFieldOf x) (getFieldOf y) id
-          where getFieldOf = appE (varE fieldName) . varE
+          ((,) fieldName <$>) <$> genApField unsafely fieldType (getFieldOf x fieldName) (getFieldOf y fieldName) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [x `asP` recP name [], varP y] body []
-genApClause unsafely (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
+   (,) constraints <$> TH.clause [x `TH.asP` TH.recP name [], varP y] body []
+genApClause unsafely instanceType (GadtC [name] fieldTypes _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genApClause unsafely (NormalC name fieldTypes)
-genApClause unsafely (RecGadtC [name] fields _resultType@(AppT _ (VarT tyVar))) =
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genApClause unsafely instanceType (NormalC name fieldTypes)
+genApClause unsafely instanceType (RecGadtC [name] fields _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genApClause unsafely (RecC name fields)
-genApClause unsafely (ForallC _vars _cxt con) = genApClause unsafely con
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genApClause unsafely instanceType (RecC name fields)
+genApClause unsafely instanceType (ForallC _vars _cxt con) = genApClause unsafely instanceType con
 
 genApField :: Bool -> Type -> Q Exp -> Q Exp -> (Q Exp -> Q Exp) -> Q ([Type], Exp)
 genApField unsafely fieldType field1Access field2Access wrap = do
@@ -347,12 +396,12 @@ genApField unsafely fieldType field1Access field2Access wrap = do
 genPureClause :: Con -> Q ([Type], Clause)
 genPureClause (NormalC name fieldTypes) = do
    argName <- newName "f"
-   let body = normalB $ appsE $ conE name : ((snd <$>) <$> constraintsAndFields)
+   let body = normalB $ TH.appsE $ conE name : ((snd <$>) <$> constraintsAndFields)
        constraintsAndFields = map newField fieldTypes
        newField :: BangType -> Q ([Type], Exp)
        newField (_, fieldType) = genPureField fieldType (varE argName) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP argName] body []
+   (,) constraints <$> TH.clause [varP argName] body []
 genPureClause (RecC name fields) = do
    argName <- newName "f"
    let body = normalB $ recConE name $ (snd <$>) <$> constraintsAndFields
@@ -360,7 +409,7 @@ genPureClause (RecC name fields) = do
        newNamedField :: VarBangType -> Q ([Type], (Name, Exp))
        newNamedField (fieldName, _, fieldType) = ((,) fieldName <$>) <$> genPureField fieldType (varE argName) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP argName] body []
+   (,) constraints <$> TH.clause [varP argName] body []
 
 genPureField :: Type -> Q Exp -> (Q Exp -> Q Exp) -> Q ([Type], Exp)
 genPureField fieldType pureValue wrap = do
@@ -373,8 +422,8 @@ genPureField fieldType pureValue wrap = do
      ParensT ty -> genPureField ty pureValue wrap
      _ -> error ("Cannot create a pure field of type " <> show fieldType)
 
-genFoldMapClause :: Con -> Q ([Type], Clause)
-genFoldMapClause (NormalC name fieldTypes) = do
+genFoldMapClause :: Type -> Con -> Q ([Type], Clause)
+genFoldMapClause _ (NormalC name fieldTypes) = do
    f          <- newName "f"
    fieldNames <- replicateM (length fieldTypes) (newName "x")
    let pats = [varP f, conP name (map varP fieldNames)]
@@ -385,8 +434,8 @@ genFoldMapClause (NormalC name fieldTypes) = do
        newField :: Name -> BangType -> Q ([Type], Exp)
        newField x (_, fieldType) = genFoldMapField f fieldType (varE x) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause pats (normalB body) []
-genFoldMapClause (RecC name fields) = do
+   (,) constraints <$> TH.clause pats (normalB body) []
+genFoldMapClause _ (RecC name fields) = do
    f <- newName "f"
    x <- newName "x"
    let body | null fields = [| mempty |]
@@ -394,18 +443,20 @@ genFoldMapClause (RecC name fields) = do
        constraintsAndFields = map newField fields
        append a b = [| $(a) <> $(b) |]
        newField :: VarBangType -> Q ([Type], Exp)
-       newField (fieldName, _, fieldType) = genFoldMapField f fieldType (appE (varE fieldName) (varE x)) id
+       newField (fieldName, _, fieldType) = genFoldMapField f fieldType (getFieldOf x fieldName) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP f, x `asP` recP name []] (normalB body) []
-genFoldMapClause (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
+   (,) constraints <$> TH.clause [varP f, x `TH.asP` TH.recP name []] (normalB body) []
+genFoldMapClause instanceType (GadtC [name] fieldTypes _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genFoldMapClause (NormalC name fieldTypes)
-genFoldMapClause (RecGadtC [name] fields _resultType@(AppT _ (VarT tyVar))) =
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genFoldMapClause instanceType (NormalC name fieldTypes)
+genFoldMapClause instanceType (RecGadtC [name] fields _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genFoldMapClause (RecC name fields)
-genFoldMapClause (ForallC _vars _cxt con) = genFoldMapClause con
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genFoldMapClause instanceType (RecC name fields)
+genFoldMapClause instanceType (ForallC _vars _cxt con) = genFoldMapClause instanceType con
 
 genFoldMapField :: Name -> Type -> Q Exp -> (Q Exp -> Q Exp) -> Q ([Type], Exp)
 genFoldMapField funcName fieldType fieldAccess wrap = do
@@ -419,10 +470,10 @@ genFoldMapField funcName fieldType fieldAccess wrap = do
      ParensT ty -> genFoldMapField funcName ty fieldAccess wrap
      _ -> (,) [] <$> [| mempty |]
 
-genTraverseClause :: Con -> Q ([Type], Clause)
-genTraverseClause (NormalC name []) =
-   (,) [] <$> clause [wildP, conP name []] (normalB [| pure $(conE name) |]) []
-genTraverseClause (NormalC name fieldTypes) = do
+genTraverseClause :: Type -> Con -> Q ([Type], Clause)
+genTraverseClause _ (NormalC name []) =
+   (,) [] <$> TH.clause [wildP, conP name []] (normalB [| pure $(conE name) |]) []
+genTraverseClause _ (NormalC name fieldTypes) = do
    f          <- newName "f"
    fieldNames <- replicateM (length fieldTypes) (newName "x")
    let pats = [varP f, conP name (map varP fieldNames)]
@@ -434,8 +485,8 @@ genTraverseClause (NormalC name fieldTypes) = do
        newField :: Name -> BangType -> Q ([Type], Exp)
        newField x (_, fieldType) = genTraverseField (varE f) fieldType (varE x) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause pats body []
-genTraverseClause (RecC name fields) = do
+   (,) constraints <$> TH.clause pats body []
+genTraverseClause _ (RecC name fields) = do
    f <- newName "f"
    x <- newName "x"
    let constraintsAndFields = map newField fields
@@ -443,18 +494,20 @@ genTraverseClause (RecC name fields) = do
        apply (a, False) b = ([| $(a) <$> $(b) |], True)
        apply (a, True) b = ([| $(a) <*> $(b) |], True)
        newField :: VarBangType -> Q ([Type], Exp)
-       newField (fieldName, _, fieldType) = genTraverseField (varE f) fieldType (appE (varE fieldName) (varE x)) id
+       newField (fieldName, _, fieldType) = genTraverseField (varE f) fieldType (getFieldOf x fieldName) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP f, x `asP` recP name []] body []
-genTraverseClause (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
+   (,) constraints <$> TH.clause [varP f, x `TH.asP` TH.recP name []] body []
+genTraverseClause instanceType (GadtC [name] fieldTypes _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genTraverseClause (NormalC name fieldTypes)
-genTraverseClause (RecGadtC [name] fields _resultType@(AppT _ (VarT tyVar))) =
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genTraverseClause instanceType (NormalC name fieldTypes)
+genTraverseClause instanceType (RecGadtC [name] fields _resultType@(AppT initType (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
-      genTraverseClause (RecC name fields)
-genTraverseClause (ForallC _vars _cxt con) = genTraverseClause con
+      let AppT classType t = instanceType
+      first (renameConstraintVars t initType <$>) <$> genTraverseClause instanceType (RecC name fields)
+genTraverseClause instanceType (ForallC _vars _cxt con) = genTraverseClause instanceType con
 
 genTraverseField :: Q Exp -> Type -> Q Exp -> (Q Exp -> Q Exp) -> Q ([Type], Exp)
 genTraverseField fun fieldType fieldAccess wrap = do
@@ -478,9 +531,9 @@ genCotraverseClause (RecC name fields) = do
        newNamedField :: VarBangType -> Q ([Type], (Name, Exp))
        newNamedField (fieldName, _, fieldType) =
           ((,) fieldName <$>) <$> (genCotraverseField ''Rank2.Distributive (varE 'Rank2.cotraverse) (varE withName)
-                                   fieldType [| $(varE fieldName) <$> $(varE argName) |] id)
+                                   fieldType [| $(projectField fieldName) <$> $(varE argName) |] id)
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP withName, varP argName] body []
+   (,) constraints <$> TH.clause [varP withName, varP argName] body []
 
 genCotraverseTraversableClause :: Con -> Q ([Type], Clause)
 genCotraverseTraversableClause (NormalC name []) = genCotraverseTraversableClause (RecC name [])
@@ -493,9 +546,28 @@ genCotraverseTraversableClause (RecC name fields) = do
        newNamedField (fieldName, _, fieldType) =
           ((,) fieldName <$>) <$> (genCotraverseField ''Rank2.DistributiveTraversable
                                    (varE 'Rank2.cotraverseTraversable) (varE withName) fieldType
-                                   [| $(varE fieldName) <$> $(varE argName) |] id)
+                                   [| $(projectField fieldName) <$> $(varE argName) |] id)
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP withName, varP argName] body []
+   (,) constraints <$> TH.clause [varP withName, varP argName] body []
+
+genDeliverClause :: TypeQ -> Maybe Name -> Con -> Q ([Type], Clause)
+genDeliverClause recType typeVar (NormalC name []) = genDeliverClause recType typeVar (RecC name [])
+genDeliverClause recType typeVar (RecC name fields) = do
+   argName <- newName "f"
+   let constraintsAndFields = map newNamedField fields
+       body = normalB $ recConE name $ (snd <$>) <$> constraintsAndFields
+       recExp g = maybe g (\v-> [|($g :: $(recType) $(varT v))|]) typeVar
+       newNamedField :: VarBangType -> Q ([Type], (Name, Exp))
+       newNamedField (fieldName, _, fieldType) =
+          ((,) fieldName <$>)
+          <$> (genDeliverField ''Rank2.Logistic fieldType
+               (\wrap-> [| \set g-> $(TH.recUpdE (recExp [|g|]) [(,) fieldName <$> appE (wrap [| Rank2.apply set |]) (getFieldOfE [|g|] fieldName)]) |])
+               (\wrap-> [| \set g-> $(TH.recUpdE (recExp [|g|]) [(,) fieldName <$> appE (wrap [| set |]) (getFieldOfE [|g|] fieldName)]) |])
+               (varE argName)
+               id
+               id)
+   constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
+   (,) constraints <$> TH.clause [varP argName] body []
 
 genCotraverseField :: Name -> Q Exp -> Q Exp -> Type -> Q Exp -> (Q Exp -> Q Exp) -> Q ([Type], Exp)
 genCotraverseField className method fun fieldType fieldAccess wrap = do
@@ -508,6 +580,77 @@ genCotraverseField className method fun fieldType fieldAccess wrap = do
      SigT ty _kind -> genCotraverseField className method fun ty fieldAccess wrap
      ParensT ty -> genCotraverseField className method fun ty fieldAccess wrap
 
+genDeliverField :: Name
+                -> Type
+                -> ((Q Exp -> Q Exp) -> Q Exp)
+                -> ((Q Exp -> Q Exp) -> Q Exp)
+                -> Q Exp
+                -> (Q Exp -> Q Exp)
+                -> (Q Exp -> Q Exp)
+                -> Q ([Type], Exp)
+genDeliverField className fieldType fieldUpdate subRecordUpdate arg outer inner = do
+   Just (Deriving _ typeVar) <- getQ
+   case fieldType of
+     AppT ty _ | ty == VarT typeVar -> (,) [] <$> outer (appE [|Compose|] ([|contramap|] `appE` fieldUpdate inner `appE` arg))
+     AppT t1 t2 | t2 == VarT typeVar ->
+                  (,) (constrain className t1) <$> outer (appE [| Rank2.deliver |] ([|contramap|] `appE` subRecordUpdate inner `appE` arg))
+     AppT t1 t2 | t1 /= VarT typeVar ->
+                  genDeliverField className t2 fieldUpdate subRecordUpdate arg (outer . appE (varE 'pure)) (inner . appE (varE 'fmap))
+     SigT ty _kind -> genDeliverField className ty fieldUpdate subRecordUpdate arg outer inner
+     ParensT ty -> genDeliverField className ty fieldUpdate subRecordUpdate arg outer inner
+
+renameConstraintVars :: Type -> Type -> Type -> Type
+renameConstraintVars (AppT instanceType (VarT instanceVar)) (AppT returnType (VarT returnVar)) constrainedType =
+   renameConstraintVars instanceType returnType (renameConstraintVar returnVar instanceVar constrainedType)
+renameConstraintVars (AppT instanceType _) (AppT returnType _) constrainedType =
+   renameConstraintVars instanceType returnType constrainedType
+renameConstraintVars _ _ constrainedType = constrainedType
+
+renameConstraintVar :: Name -> Name -> Type -> Type
+renameConstraintVar from to (VarT name)
+   | name == from = VarT to
+   | otherwise = VarT name
+renameConstraintVar from to (AppT a b) = AppT (renameConstraintVar from to a) (renameConstraintVar from to b)
+renameConstraintVar from to (AppKindT t k) = AppT (renameConstraintVar from to t) (renameConstraintVar from to k)
+renameConstraintVar from to (InfixT a op b) = InfixT (renameConstraintVar from to a) op (renameConstraintVar from to b)
+renameConstraintVar from to (UInfixT a op b) = UInfixT (renameConstraintVar from to a) op (renameConstraintVar from to b)
+renameConstraintVar from to (SigT t k) = SigT (renameConstraintVar from to t) (renameConstraintVar from to k)
+renameConstraintVar from to (ParensT t) = ParensT (renameConstraintVar from to t)
+renameConstraintVar _ _ t = t
+
+projectField :: Name -> Q Exp
+projectField field = do
+#if MIN_VERSION_template_haskell(2,19,0)
+  dotty <- TH.isExtEnabled TH.OverloadedRecordDot
+  if dotty
+     then TH.projectionE (pure $ TH.nameBase field)
+     else varE field
+#else
+  varE field
+#endif
+
+getFieldOf :: Name -> Name -> Q Exp
+getFieldOf = getFieldOfE . varE
+
+getFieldOfE :: Q Exp -> Name -> Q Exp
+getFieldOfE record field = do
+#if MIN_VERSION_template_haskell(2,19,0)
+  dotty <- TH.isExtEnabled TH.OverloadedRecordDot
+  if dotty
+     then TH.getFieldE record (TH.nameBase field)
+     else appE (varE field) record
+#else
+  appE (varE field) record
+#endif
+
 constrain :: Name -> Type -> [Type]
 constrain _ ConT{} = []
 constrain cls t = [ConT cls `AppT` t]
+
+#if MIN_VERSION_template_haskell(2,17,0)
+binder :: Name -> TyVarBndr TH.Specificity
+binder name = TH.PlainTV name TH.SpecifiedSpec
+#else
+binder :: Name -> TyVarBndr
+binder = TH.PlainTV
+#endif
